#ifndef _GI_GDKPIXBUF_GDKPIXBUF_SETUP_PRE_DEF_HPP_
#define _GI_GDKPIXBUF_GDKPIXBUF_SETUP_PRE_DEF_HPP_

/* some header parts are only conditionally considered */
#ifndef GDK_PIXBUF_ENABLE_BACKEND
#define GDK_PIXBUF_ENABLE_BACKEND
#endif

#endif
